// SPDX-FileCopyrightText: 2011-2019 Disney Enterprises, Inc.
// SPDX-License-Identifier: LicenseRef-Apache-2.0
// SPDX-FileCopyrightText: 2020 L. E. Segovia <amy@amyspark.me>
// SPDX-License-Identifier: GPL-3.0-or-later

#pragma once

/* #undef SEEXPR_ENABLE_LLVM */

#define KSeExpr_HAVE_CHARCONV_WITH_DOUBLES

/* #undef KSeExpr_HAVE_DYNAMIC_DISPATCH */
